from jax import Array
from flax import nnx

from soma_models.v1.configs import EncoderConfig, LayerConfig
from soma_models.v1.flax.modules.layer import Layer


class Encoder(nnx.Module):
    def __init__(self, config: EncoderConfig, rngs: nnx.Rngs):
        self.layers = nnx.List(
            [
                Layer(
                    LayerConfig(
                        dropout_rate=config.dropout_rate,
                        embedding_dim=config.embedding_dim,
                        pwff_hidden_dim=config.pwff_hidden_dim,
                        num_heads=config.num_heads,
                        max_wavelength=config.max_wavelength,
                        scale_factor=config.scale_factor,
                    ),
                    rngs=rngs,
                )
                for _ in range(config.num_layers)
            ]
        )

    def __call__(
        self,
        input: Array,
        positions: Array,
    ):
        x = input
        for layer in self.layers:
            x = layer(x, positions)
        return x



import torch
import torch.nn.functional as F

from soma_models.v1.configs import V1_PAD_TOKEN_ID
from soma_models.v1.torch.modules.model import Model
from soma_models.v1.torch.modules.sig_reg import SIGReg


def compute_loss(
    model: Model,
    sig_reg: SIGReg,
    token_ids: torch.Tensor,
    targets: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Compute the training loss (cross-entropy + SIGReg regularization).

    Args:
        model: The V1 Model instance.
        sig_reg: The SIGReg module instance (generates noise internally via global RNG).
        token_ids: Input token ids, shape [batch, seq].
        targets: Next-token targets, shape [batch, seq].

    Returns:
        (loss, embedding) where:
            loss: Scalar tensor = cross_entropy + sig_reg_loss
            embedding: Mean embedding, shape [embedding_dim]
    """
    batch_size, seq_len = token_ids.shape
    positions = torch.arange(seq_len, device=token_ids.device).unsqueeze(0).expand(batch_size, -1)

    representations = model.encode(token_ids, positions)
    logits = model.predict(representations)
    sig_reg_loss = sig_reg(representations)

    embedding = representations.mean(dim=1).mean(dim=0)

    logits_flat = logits.reshape(-1, logits.shape[-1])
    targets_flat = targets.reshape(-1)
    ce_loss = F.cross_entropy(logits_flat, targets_flat.long(), ignore_index=V1_PAD_TOKEN_ID)

    return ce_loss + sig_reg_loss, embedding



from dataclasses import dataclass

V1_PAD_TOKEN_ID = 256
V1_EOS_TOKEN_ID = 257

V1_EMBEDDING_DIM = 2048
V1_NUM_HEADS = 8
V1_NUM_LAYERS = 32
V1_MAX_SEQ_LEN = 8192
V1_PWFF_HIDDEN_DIM = V1_EMBEDDING_DIM * 4
V1_MAX_WAVELENGTH = 10_000.0
V1_SCALE_FACTOR = 1.0
V1_VOCAB_SIZE = 256 + 8

V1_SIG_REG_T_MAX = 3.0
V1_SIG_REG_SLICES = 1024
V1_SIG_REG_POINTS = 17
V1_SIG_REG_COEFFICIENT = 0.02

V1_BATCH_SIZE = 32


@dataclass
class ModelConfig:
    dropout_rate: float
    embedding_dim: int = V1_EMBEDDING_DIM
    pwff_hidden_dim: int = V1_PWFF_HIDDEN_DIM
    num_layers: int = V1_NUM_LAYERS
    num_heads: int = V1_NUM_HEADS
    vocab_size: int = V1_VOCAB_SIZE
    max_wavelength: float = V1_MAX_WAVELENGTH
    scale_factor: float = V1_SCALE_FACTOR


@dataclass
class EncoderConfig:
    dropout_rate: float
    embedding_dim: int = V1_EMBEDDING_DIM
    pwff_hidden_dim: int = V1_PWFF_HIDDEN_DIM
    num_layers: int = V1_NUM_LAYERS
    num_heads: int = V1_NUM_HEADS
    max_wavelength: float = V1_MAX_WAVELENGTH
    scale_factor: float = V1_SCALE_FACTOR


@dataclass
class LayerConfig:
    dropout_rate: float
    embedding_dim: int = V1_EMBEDDING_DIM
    pwff_hidden_dim: int = V1_PWFF_HIDDEN_DIM
    num_heads: int = V1_NUM_HEADS
    max_wavelength: float = V1_MAX_WAVELENGTH
    scale_factor: float = V1_SCALE_FACTOR


@dataclass
class PositionWiseFeedForwardConfig:
    dropout_rate: float
    embedding_dim: int = V1_EMBEDDING_DIM
    pwff_hidden_dim: int = V1_PWFF_HIDDEN_DIM


@dataclass
class SIGRegConfig:
    t_max: float = V1_SIG_REG_T_MAX
    points: int = V1_SIG_REG_POINTS
    slices: int = V1_SIG_REG_SLICES
    coefficient: float = V1_SIG_REG_COEFFICIENT



from jax import Array
from flax import nnx

from soma_models.v1.configs import LayerConfig, PositionWiseFeedForwardConfig
from soma_models.v1.flax.modules.pwff import PositionWiseFeedForward
from soma_models.v1.flax.modules.attention import MultiHeadAttention


class Layer(nnx.Module):
    def __init__(self, config: LayerConfig, rngs: nnx.Rngs):
        self.norm_1 = nnx.LayerNorm(
            num_features=config.embedding_dim,
            rngs=rngs,
            epsilon=1e-5,
            use_fast_variance=False,
        )
        self.attention = MultiHeadAttention(
            num_heads=config.num_heads,
            num_features=config.embedding_dim,
            dropout_rate=config.dropout_rate,
            max_wavelength=config.max_wavelength,
            scale_factor=config.scale_factor,
            rngs=rngs,
        )
        self.norm_2 = nnx.LayerNorm(
            num_features=config.embedding_dim,
            rngs=rngs,
            use_fast_variance=False,
            epsilon=1e-5,
        )
        self.pwff = PositionWiseFeedForward(
            PositionWiseFeedForwardConfig(
                dropout_rate=config.dropout_rate,
                embedding_dim=config.embedding_dim,
                pwff_hidden_dim=config.pwff_hidden_dim,
            ),
            rngs,
        )
        self.dropout = nnx.Dropout(config.dropout_rate, rngs=rngs)

    def __call__(self, context: Array, positions: Array):
        x = context
        residual_path = self.norm_1(x)
        residual_path = self.attention(
            residual_path, positions=positions
        )
        residual_path = self.dropout(residual_path)
        x = x + residual_path
        residual_path = self.norm_2(x)
        residual_path = self.pwff(residual_path)
        residual_path = self.dropout(residual_path)
        x = x + residual_path
        return x



import jax
import jax.numpy as jnp
from flax import nnx
from jax import Array

from soma_models.v1.configs import SIGRegConfig


class SIGReg(nnx.Module):
    def __init__(self, config: SIGRegConfig, rngs: nnx.Rngs):
        self.config = config
        self.rngs = rngs
        t = jnp.linspace(0, config.t_max, config.points)
        dt = config.t_max / (config.points - 1)
        weights = jnp.full((config.points,), 2 * dt)
        weights = weights.at[0].set(dt)
        weights = weights.at[-1].set(dt)
        self.t = t
        self.phi = jnp.exp(-0.5 * (t**2))
        self.weights = weights * self.phi

    def __call__(self, x: Array) -> Array:
        noise = jax.random.normal(
            self.rngs.noise(), shape=(x.shape[-1], self.config.slices), dtype=x.dtype
        )
        return self.compute(x, noise)

    def compute(self, x: Array, noise: Array) -> Array:
        A = noise / jnp.linalg.norm(noise, axis=0)
        x = x @ A
        N = x.shape[-1]
        x_t = x[..., None] * self.t
        cos_mean = jnp.mean(jnp.cos(x_t), axis=-2)
        sin_mean = jnp.mean(jnp.sin(x_t), axis=-2)
        err = ((cos_mean - self.phi) ** 2) + (sin_mean**2)
        integrated = jnp.sum(err * self.weights, axis=-1) * N
        return integrated.mean() * self.config.coefficient



import torch
import torch.nn as nn
import torch.nn.functional as F
import math


def apply_rope(
    inputs: torch.Tensor,  # [BATCH_SIZE, SEQ_LEN, NUM_HEADS, HEAD_DIM]
    positions: torch.Tensor,  # [BATCH_SIZE, SEQ_LEN]
    head_dim: int,
    max_wavelength: int = 10_000,
    scale_factor: float = 1.0,
) -> torch.Tensor:
    fraction = 2 * torch.arange(0, head_dim // 2, dtype=inputs.dtype, device=inputs.device) / head_dim
    timescale = max_wavelength ** fraction
    positions = positions.unsqueeze(-1).to(inputs.dtype)  # [batch, seq, 1]
    timescale = timescale.unsqueeze(0).unsqueeze(0)  # [1, 1, head_dim//2]

    sinusoid_inp = positions / timescale
    sinusoid_inp = sinusoid_inp.unsqueeze(-2)  # [batch, seq, 1, head_dim//2]
    if scale_factor < 1.0:
        raise ValueError(f"scale_factor must be >= 1.0, got {scale_factor}")
    sinusoid_inp = sinusoid_inp / scale_factor

    sin = torch.sin(sinusoid_inp)
    cos = torch.cos(sinusoid_inp)

    first_half, second_half = inputs.chunk(2, dim=-1)
    first_part = first_half * cos - second_half * sin
    second_part = second_half * cos + first_half * sin
    out = torch.cat([first_part, second_part], dim=-1)
    return out.to(inputs.dtype)


class MultiHeadAttention(nn.Module):
    def __init__(
        self,
        num_heads: int,
        num_features: int,
        *,
        dropout_rate: float = 0.0,
        use_bias: bool = True,
        max_wavelength: float = 10_000.0,
        scale_factor: float = 1.0,
    ):
        super().__init__()
        self.num_heads = num_heads
        self.num_features = num_features
        self.dropout_rate = dropout_rate
        self.max_wavelength = max_wavelength
        self.scale_factor = scale_factor

        if num_features % num_heads != 0:
            raise ValueError(
                f"Memory dimension ({num_features}) must be divisible by "
                f"'num_heads' heads ({num_heads})."
            )
        self.head_dim = num_features // num_heads

        self.query = nn.Linear(num_features, num_features, bias=use_bias)
        self.key = nn.Linear(num_features, num_features, bias=use_bias)
        self.value = nn.Linear(num_features, num_features, bias=use_bias)
        self.output = nn.Linear(num_features, num_features, bias=use_bias)

    def forward(
        self,
        inputs: torch.Tensor,
        *,
        positions: torch.Tensor | None = None,
    ):
        batch_size, seq_len, _ = inputs.shape

        # Project to Q, K, V and reshape to [batch, seq, heads, head_dim]
        query = self.query(inputs).reshape(batch_size, seq_len, self.num_heads, self.head_dim)
        key = self.key(inputs).reshape(batch_size, seq_len, self.num_heads, self.head_dim)
        value = self.value(inputs).reshape(batch_size, seq_len, self.num_heads, self.head_dim)

        if positions is not None:
            query = apply_rope(query, positions, head_dim=self.head_dim, max_wavelength=self.max_wavelength, scale_factor=self.scale_factor)
            key = apply_rope(key, positions, head_dim=self.head_dim, max_wavelength=self.max_wavelength, scale_factor=self.scale_factor)

        # Transpose to [batch, heads, seq, head_dim] for attention
        query = query.transpose(1, 2)
        key = key.transpose(1, 2)
        value = value.transpose(1, 2)

        # Scale
        scale = math.sqrt(self.head_dim)
        attn_weights = torch.matmul(query, key.transpose(-2, -1)) / scale

        # Generate causal mask: [batch, 1, seq, seq] with True=attend, False=mask
        mask = torch.ones(batch_size, 1, seq_len, seq_len, dtype=torch.bool, device=inputs.device).tril()
        big_neg = torch.finfo(attn_weights.dtype).min
        attn_weights = torch.where(mask, attn_weights, big_neg)

        attn_weights = F.softmax(attn_weights, dim=-1)

        if self.training and self.dropout_rate > 0:
            attn_weights = F.dropout(attn_weights, p=self.dropout_rate)

        # [batch, heads, seq, head_dim]
        attn_output = torch.matmul(attn_weights, value)

        # Transpose back to [batch, seq, heads, head_dim] and reshape
        attn_output = attn_output.transpose(1, 2).reshape(batch_size, seq_len, self.num_features)

        out = self.output(attn_output)
        return out



import jax
import jax.numpy as jnp
from jax import Array

from soma_models.v1.configs import V1_PAD_TOKEN_ID
from soma_models.v1.flax.modules.model import Model
from soma_models.v1.flax.modules.sig_reg import SIGReg


def compute_loss(
    model: Model,
    sig_reg: SIGReg,
    token_ids: Array,
    targets: Array,
) -> tuple[Array, Array]:
    """Compute the training loss (cross-entropy + SIGReg regularization).

    Args:
        model: The V1 Model instance.
        sig_reg: The SIGReg module instance (generates noise internally via rngs).
        token_ids: Input token ids, shape [batch, seq].
        targets: Next-token targets, shape [batch, seq].

    Returns:
        (loss, embedding) where:
            loss: Scalar = cross_entropy + sig_reg_loss
            embedding: Mean embedding, shape [embedding_dim]
    """
    batch_size, seq_len = token_ids.shape
    positions = jnp.broadcast_to(jnp.arange(seq_len), (batch_size, seq_len))

    representations = model.encode(token_ids, positions)
    logits = model.predict(representations)
    sig_reg_loss = sig_reg(representations)

    embedding = jnp.mean(jnp.mean(representations, axis=1), axis=0)

    logits_flat = logits.reshape(-1, logits.shape[-1])
    targets_flat = targets.reshape(-1)
    mask = targets_flat != V1_PAD_TOKEN_ID

    log_probs = jax.nn.log_softmax(logits_flat, axis=-1)
    ce_per_token = -log_probs[jnp.arange(logits_flat.shape[0]), targets_flat]
    ce_loss = jnp.sum(ce_per_token * mask) / jnp.maximum(jnp.sum(mask), 1)

    return ce_loss + sig_reg_loss, embedding



def flatten_dict(nested_dict, parent_key="", sep="."):
    items = []
    for k, v in nested_dict.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict) and v:  # Check if v is a non-empty dictionary
            # Recursively flatten the nested dictionary
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            # If v is not a dictionary or an empty dictionary, add it as is
            items.append((new_key, v))
    return dict(items)


def unflatten_dict(flat_dict, sep="."):
    result_dict = {}
    for key, value in flat_dict.items():
        parts = key.split(sep)
        d = result_dict
        for i, part in enumerate(
            parts[:-1]
        ):  # Iterate through all parts except the last one
            if part not in d:
                d[part] = {}
            elif not isinstance(d[part], dict):
                raise ValueError(
                    f"Unflattening conflict: key '{part}' in path '{key}' "
                    f"already exists with a non-dictionary value: {d[part]}"
                )
            d = d[part]

        last_part = parts[-1]
        if (
            last_part in d
            and isinstance(d[last_part], dict)
            and not isinstance(value, dict)
        ):
            raise ValueError(
                f"Unflattening conflict: key '{last_part}' in path '{key}' "
                f"is already a dictionary, but trying to assign a non-dictionary value: {value}"
            )

        d[last_part] = value
    return result_dict



from soma_models.v1.configs import ModelConfig, SIGRegConfig
from soma_models.v1.torch.modules.model import Model
from soma_models.v1.torch.modules.sig_reg import SIGReg
from soma_models.v1.torch.loss import compute_loss

__all__ = ["Model", "ModelConfig", "SIGReg", "SIGRegConfig", "compute_loss"]



import os
from typing import Dict, Union, List, TypeVar, Generic, Tuple
from soma_models.utils import remap, flatten_dict, unflatten_dict
from safetensors.flax import save_file, load_file, save, load, Array
from flax import nnx

M = TypeVar("M", bound=nnx.Module)

# (safetensor_suffix, flax_suffix) pairs per module type
_LAYER_MAPPINGS: Dict[type, List[Tuple[str, str]]] = {
    nnx.Embed: [("weight", "embedding")],
    nnx.Linear: [("weight", "kernel")],
    nnx.LayerNorm: [("gamma", "scale"), ("beta", "bias")],
}

_MHA_MAPPINGS: List[Tuple[str, str]] = [
    ("query.weight", "query.kernel"),
    ("key.weight", "key.kernel"),
    ("value.weight", "value.kernel"),
    ("output.weight", "out.kernel"),
    ("output.bias", "out.bias"),
]


class Serde(Generic[M]):
    def __init__(self, module: M):
        self.module = module
        self.attention_shapes: Dict[str, Tuple[int, int]] = {}
        self.map_to_flax: Dict[str, str] = {}
        self.map_to_safetensor: Dict[str, str] = {}
        self.remove: List[str] = []

        for path, submodule in nnx.iter_modules(module):
            prefix = ".".join(str(item) for item in path)

            if isinstance(submodule, nnx.Dropout):
                continue

            from soma_models.v1.flax.modules.attention import MultiHeadAttention
            if isinstance(submodule, MultiHeadAttention):
                self.attention_shapes[prefix] = (
                    submodule.num_heads,
                    submodule.head_dim,
                )
                for st_suffix, flax_suffix in _MHA_MAPPINGS:
                    self._add(f"{prefix}.{st_suffix}", f"{prefix}.{flax_suffix}")
                continue

            mappings = _LAYER_MAPPINGS.get(type(submodule))
            if mappings:
                for st_suffix, flax_suffix in mappings:
                    self._add(f"{prefix}.{st_suffix}", f"{prefix}.{flax_suffix}")

    def _add(self, safetensor_key: str, flax_key: str):
        self.map_to_flax[safetensor_key] = flax_key
        self.map_to_safetensor[flax_key] = safetensor_key

    def _reshape_attention(
        self,
        params: Dict[str, Array],
        name: str,
        num_heads: int,
        head_dim: int,
        *,
        to_flax: bool,
    ):
        """Reshape attention parameters between flat (safetensor) and split-head (flax) layouts."""
        num_features = num_heads * head_dim
        flat = (num_features, num_features)
        # Key names differ by direction: safetensor uses "weight"/"output", flax uses "kernel"/"out"
        if to_flax:
            qkv_key = "weight"
            out_w_key = f"{name}.output.weight"
            out_b_key = f"{name}.output.bias"
        else:
            qkv_key = "kernel"
            out_w_key = f"{name}.out.kernel"
            out_b_key = f"{name}.out.bias"

        for component in ["query", "key", "value"]:
            w = f"{name}.{component}.{qkv_key}"
            if w in params:
                params[w] = params[w].reshape(
                    (num_features, num_heads, head_dim) if to_flax else flat
                )
            b = f"{name}.{component}.bias"
            if b in params:
                params[b] = params[b].reshape(
                    (num_heads, head_dim) if to_flax else (num_features,)
                )

        if out_w_key in params:
            params[out_w_key] = params[out_w_key].reshape(
                (num_heads, head_dim, num_features) if to_flax else flat
            )

    def _serialize_common(self) -> Dict:
        state = nnx.state(self.module)
        state_dict = nnx.to_pure_dict(state)
        flat_state_dict = flatten_dict(state_dict)
        # Remove all RNG stream state (rngs.count / rngs.key) from any module
        rng_keys = [k for k in flat_state_dict if k.endswith(".rngs.count") or k.endswith(".rngs.key")]
        for k in rng_keys:
            del flat_state_dict[k]
        for name, (num_heads, head_dim) in self.attention_shapes.items():
            self._reshape_attention(
                flat_state_dict, name, num_heads, head_dim, to_flax=False
            )
        remap(flat_state_dict, self.map_to_safetensor, self.remove)
        return dict(sorted(flat_state_dict.items()))

    def serialize(self) -> bytes:
        return save(self._serialize_common())

    def serialize_to_file(self, filename: Union[str, os.PathLike]) -> None:
        return save_file(self._serialize_common(), filename)

    def _deserialize_common(self, safetensor_dict: Dict[str, Array]) -> M:
        for name, (num_heads, head_dim) in self.attention_shapes.items():
            self._reshape_attention(
                safetensor_dict, name, num_heads, head_dim, to_flax=True
            )
        remap(safetensor_dict, self.map_to_flax, [])
        nested_dict = unflatten_dict(safetensor_dict)
        state = nnx.state(self.module)
        nnx.replace_by_pure_dict(state, nested_dict)
        nnx.update(self.module, state)
        return self.module

    def deserialize(self, data: bytes) -> M:
        return self._deserialize_common(load(data))

    def deserialize_from_file(self, filename: Union[str, os.PathLike]) -> M:
        return self._deserialize_common(load_file(filename))


class Serializable:
    """Mixin for nnx.Module subclasses that adds safetensors serialization."""

    def save(self, filename: Union[str, os.PathLike]) -> None:
        Serde(self).serialize_to_file(filename)

    def save_bytes(self) -> bytes:
        return Serde(self).serialize()

    @classmethod
    def load(cls, filename: Union[str, os.PathLike], *args, **kwargs):
        module = cls(*args, **kwargs)
        Serde(module).deserialize_from_file(filename)
        return module

    @classmethod
    def load_bytes(cls, data: bytes, *args, **kwargs):
        module = cls(*args, **kwargs)
        Serde(module).deserialize(data)
        return module



import torch
import torch.nn as nn

from soma_models.v1.configs import EncoderConfig, LayerConfig
from soma_models.v1.torch.modules.layer import Layer


class Encoder(nn.Module):
    def __init__(self, config: EncoderConfig):
        super().__init__()
        self.layers = nn.ModuleList([
            Layer(LayerConfig(
                dropout_rate=config.dropout_rate,
                embedding_dim=config.embedding_dim,
                pwff_hidden_dim=config.pwff_hidden_dim,
                num_heads=config.num_heads,
                max_wavelength=config.max_wavelength,
                scale_factor=config.scale_factor,
            ))
            for _ in range(config.num_layers)
        ])

    def forward(self, input: torch.Tensor, positions: torch.Tensor):
        x = input
        for layer in self.layers:
            x = layer(x, positions)
        return x



"""Framework-agnostic byte-level tokenizer matching the on-chain V1 data contract.

Converts raw bytes into (token_ids, targets, pos_ids) batches that can be
consumed by any ML framework — wrap the returned lists with ``torch.tensor()``,
``jnp.array()``, ``tf.constant()``, etc.

The logic mirrors the Rust ``ByteSequenceDataset`` / ``ByteSequenceBatcher``
in ``models/src/v1/data/`` exactly.
"""

from __future__ import annotations

from soma_models.v1.configs import (
    V1_BATCH_SIZE,
    V1_EOS_TOKEN_ID,
    V1_MAX_SEQ_LEN,
    V1_PAD_TOKEN_ID,
)


class ByteSequenceBatch:
    """A single batch of tokenized byte sequences.

    Attributes:
        token_ids: ``[batch, seq_len]`` nested list of ints.
        targets: ``[batch, seq_len]`` nested list of ints (token_ids shifted
            left by 1, with PAD appended).
        pos_ids: ``[batch, seq_len]`` nested list of ints (global byte
            offsets; PAD/EOS positions clamped to last-data-byte + 1).
    """

    __slots__ = ("token_ids", "targets", "pos_ids")

    def __init__(
        self,
        token_ids: list[list[int]],
        targets: list[list[int]],
        pos_ids: list[list[int]],
    ) -> None:
        self.token_ids = token_ids
        self.targets = targets
        self.pos_ids = pos_ids


def tokenize(
    data: bytes | bytearray,
    max_seq_len: int = V1_MAX_SEQ_LEN,
    batch_size: int = V1_BATCH_SIZE,
) -> list[ByteSequenceBatch]:
    """Tokenize raw bytes into batches matching the on-chain V1 data contract.

    Args:
        data: Raw byte data to tokenize.
        max_seq_len: Maximum sequence length per chunk.
        batch_size: Number of sequences per batch.

    Returns:
        A list of ``ByteSequenceBatch`` instances.  The final batch may
        contain fewer than ``batch_size`` sequences (matching the Rust
        DataLoader behaviour).
    """
    if len(data) == 0:
        return []

    num_chunks = -(-len(data) // max_seq_len)  # ceil division

    items: list[tuple[list[int], list[int], list[int]]] = []
    for index in range(num_chunks):
        start = index * max_seq_len
        remaining = len(data) - start
        data_len = min(remaining, max_seq_len)

        is_final = index + 1 == num_chunks
        has_room = data_len < max_seq_len
        eos_pos = data_len if (is_final and has_room) else -1

        token_ids: list[int] = []
        pos_ids: list[int] = []
        pos_after_last = start + data_len

        for i in range(max_seq_len):
            if i < data_len:
                token_ids.append(data[start + i])
                pos_ids.append(start + i)
            elif i == eos_pos:
                token_ids.append(V1_EOS_TOKEN_ID)
                pos_ids.append(pos_after_last)
            else:
                token_ids.append(V1_PAD_TOKEN_ID)
                pos_ids.append(pos_after_last)

        targets = token_ids[1:] + [V1_PAD_TOKEN_ID]
        items.append((token_ids, targets, pos_ids))

    batches: list[ByteSequenceBatch] = []
    for i in range(0, len(items), batch_size):
        batch_items = items[i : i + batch_size]
        batch_token_ids = [item[0] for item in batch_items]
        batch_targets = [item[1] for item in batch_items]
        batch_pos_ids = [item[2] for item in batch_items]
        batches.append(ByteSequenceBatch(batch_token_ids, batch_targets, batch_pos_ids))

    return batches



import torch
import torch.nn as nn

from soma_models.v1.configs import LayerConfig, PositionWiseFeedForwardConfig
from soma_models.v1.torch.modules.pwff import PositionWiseFeedForward
from soma_models.v1.torch.modules.attention import MultiHeadAttention


class Layer(nn.Module):
    def __init__(self, config: LayerConfig):
        super().__init__()
        self.norm_1 = nn.LayerNorm(config.embedding_dim, eps=1e-5)
        self.attention = MultiHeadAttention(
            num_heads=config.num_heads,
            num_features=config.embedding_dim,
            dropout_rate=config.dropout_rate,
            max_wavelength=config.max_wavelength,
            scale_factor=config.scale_factor,
        )
        self.norm_2 = nn.LayerNorm(config.embedding_dim, eps=1e-5)
        self.pwff = PositionWiseFeedForward(
            PositionWiseFeedForwardConfig(
                dropout_rate=config.dropout_rate,
                embedding_dim=config.embedding_dim,
                pwff_hidden_dim=config.pwff_hidden_dim,
            )
        )
        self.dropout = nn.Dropout(config.dropout_rate)

    def forward(self, context: torch.Tensor, positions: torch.Tensor):
        x = context
        residual_path = self.norm_1(x)
        residual_path = self.attention(residual_path, positions=positions)
        residual_path = self.dropout(residual_path)
        x = x + residual_path
        residual_path = self.norm_2(x)
        residual_path = self.pwff(residual_path)
        residual_path = self.dropout(residual_path)
        x = x + residual_path
        return x



from typing import Dict, Any, List


def remap(
    d: Dict[str, Any], rename_map: Dict[str, str], keys_to_delete: List[str]
) -> None:
    """Rename keys in `d` according to `rename_map` and delete `keys_to_delete`, in-place."""
    for key in keys_to_delete:
        d.pop(key, None)

    for old_key, new_key in rename_map.items():
        if old_key in d and old_key != new_key:
            d[new_key] = d.pop(old_key)



import torch.nn as nn
import torch.nn.functional as F

from soma_models.v1.configs import PositionWiseFeedForwardConfig


class PositionWiseFeedForward(nn.Module):
    def __init__(self, config: PositionWiseFeedForwardConfig):
        super().__init__()
        self.linear_inner = nn.Linear(config.embedding_dim, config.pwff_hidden_dim)
        self.linear_outer = nn.Linear(config.pwff_hidden_dim, config.embedding_dim)
        self.dropout = nn.Dropout(config.dropout_rate)

    def forward(self, x):
        x = self.linear_inner(x)
        x = F.gelu(x, approximate='none')
        x = self.dropout(x)
        output = self.linear_outer(x)
        return output



import torch
import torch.nn as nn

from soma_models.v1.configs import ModelConfig, EncoderConfig
from soma_models.v1.torch.modules.encoder import Encoder
from soma_models.v1.torch.serde import Serializable


class Model(Serializable, nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.embedding = nn.Embedding(config.vocab_size, config.embedding_dim)
        self.encoder = Encoder(
            EncoderConfig(
                dropout_rate=config.dropout_rate,
                embedding_dim=config.embedding_dim,
                pwff_hidden_dim=config.pwff_hidden_dim,
                num_layers=config.num_layers,
                num_heads=config.num_heads,
                max_wavelength=config.max_wavelength,
                scale_factor=config.scale_factor,
            )
        )
        self.final_norm = nn.LayerNorm(config.embedding_dim, eps=1e-5)
        self.predictor = nn.Linear(config.embedding_dim, config.vocab_size)

    def encode(
        self,
        input: torch.Tensor,
        positions: torch.Tensor,
    ) -> torch.Tensor:
        x = self.embedding(input)
        x = self.encoder(x, positions)
        x = self.final_norm(x)
        return x

    def predict(self, embeddings: torch.Tensor) -> torch.Tensor:
        return self.predictor(embeddings)

    def forward(
        self,
        input: torch.Tensor,
        positions: torch.Tensor,
    ) -> torch.Tensor:
        return self.predict(self.encode(input, positions))



import torch
import torch.nn as nn

from soma_models.v1.configs import SIGRegConfig


class SIGReg(nn.Module):
    def __init__(self, config: SIGRegConfig):
        super().__init__()
        self.config = config
        t = torch.linspace(0, config.t_max, config.points)
        dt = config.t_max / (config.points - 1)
        weights = torch.full((config.points,), 2 * dt)
        weights[0] = dt
        weights[-1] = dt
        phi = torch.exp(-0.5 * (t**2))
        self.register_buffer("t", t)
        self.register_buffer("phi", phi)
        self.register_buffer("weights", weights * phi)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        noise = torch.randn(
            x.shape[-1], self.config.slices, dtype=x.dtype, device=x.device
        )
        return self.compute(x, noise)

    def compute(self, x: torch.Tensor, noise: torch.Tensor) -> torch.Tensor:
        A = noise / torch.norm(noise, dim=0)
        x = x @ A
        N = x.shape[-1]
        x_t = x.unsqueeze(-1) * self.t
        cos_mean = torch.mean(torch.cos(x_t), dim=-2)
        sin_mean = torch.mean(torch.sin(x_t), dim=-2)
        err = ((cos_mean - self.phi) ** 2) + (sin_mean**2)
        integrated = torch.sum(err * self.weights, dim=-1) * N
        return integrated.mean() * self.config.coefficient



from flax import nnx

from soma_models.v1.configs import PositionWiseFeedForwardConfig


class PositionWiseFeedForward(nnx.Module):
    def __init__(self, config: PositionWiseFeedForwardConfig, rngs: nnx.Rngs):
        self.linear_inner = nnx.Linear(
            in_features=config.embedding_dim,
            out_features=config.pwff_hidden_dim,
            rngs=rngs,
        )
        self.linear_outer = nnx.Linear(
            in_features=config.pwff_hidden_dim,
            out_features=config.embedding_dim,
            rngs=rngs,
        )
        self.dropout = nnx.Dropout(config.dropout_rate, rngs=rngs)

    def __call__(self, x):
        x = self.linear_inner(x)
        x = nnx.gelu(x, approximate=False)
        x = self.dropout(x)
        output = self.linear_outer(x)
        return output



from jax import Array
from flax import nnx

from soma_models.v1.configs import ModelConfig, EncoderConfig
from soma_models.v1.flax.modules.encoder import Encoder
from soma_models.v1.flax.serde import Serializable


class Model(Serializable, nnx.Module):
    def __init__(self, config: ModelConfig, rngs: nnx.Rngs) -> None:
        self.config = config
        self.embedding = nnx.Embed(config.vocab_size, config.embedding_dim, rngs=rngs)
        self.encoder = Encoder(
            EncoderConfig(
                dropout_rate=config.dropout_rate,
                embedding_dim=config.embedding_dim,
                pwff_hidden_dim=config.pwff_hidden_dim,
                num_layers=config.num_layers,
                num_heads=config.num_heads,
                max_wavelength=config.max_wavelength,
                scale_factor=config.scale_factor,
            ),
            rngs,
        )
        self.final_norm = nnx.LayerNorm(
            config.embedding_dim, epsilon=1e-5, use_fast_variance=False, rngs=rngs
        )
        self.predictor = nnx.Linear(config.embedding_dim, config.vocab_size, rngs=rngs)

    def encode(
        self,
        input: Array,
        positions: Array,
    ) -> Array:
        x = self.embedding(input)
        x = self.encoder(x, positions)
        x = self.final_norm(x)
        return x

    def predict(
        self,
        embeddings: Array,
    ) -> Array:
        return self.predictor(embeddings)

    def __call__(
        self,
        input: Array,
        positions: Array,
    ) -> Array:
        return self.predict(self.encode(input, positions))



from collections.abc import Callable
import functools
from jax import Array
import jax.numpy as jnp
from flax.nnx.module import Module, first_from
from flax.nnx.nn.attention import dot_product_attention
from flax.typing import (
    Dtype,
    Initializer,
    PrecisionLike,
)
from flax.nnx.nn import initializers
from flax.nnx.nn.linear import (
    LinearGeneral,
    default_kernel_init,
)
from flax.nnx import rnglib


def apply_rope(
    inputs: Array,  # [BATCH_SIZE, SEQ_LEN, NUM_HEADS, HEAD_DIM]
    positions: Array,  # [BATCH_SIZE, SEQ_LEN]
    head_dim: int,
    max_wavelength: int = 10_000,
    scale_factor: float = 1.0,
) -> Array:
    fraction = 2 * jnp.arange(0, head_dim // 2) / head_dim
    timescale = max_wavelength**fraction
    positions = positions[..., jnp.newaxis]
    timescale = timescale[jnp.newaxis, jnp.newaxis, :]  # ty: ignore[non-subscriptable]

    sinusoid_inp = positions / timescale
    sinusoid_inp = sinusoid_inp[..., jnp.newaxis, :]
    if scale_factor < 1.0:
        raise ValueError(f"scale_factor must be >= 1.0, got {scale_factor}")
    sinusoid_inp /= scale_factor

    sin = jnp.sin(sinusoid_inp)
    cos = jnp.cos(sinusoid_inp)

    first_half, second_half = jnp.split(inputs, 2, axis=-1)
    first_part = first_half * cos - second_half * sin
    second_part = second_half * cos + first_half * sin
    out = jnp.concatenate([first_part, second_part], axis=-1)
    return out.astype(inputs.dtype)


class MultiHeadAttention(Module):
    def __init__(
        self,
        num_heads: int,
        num_features: int,
        *,
        dtype: Dtype | None = None,
        param_dtype: Dtype = jnp.float32,
        broadcast_dropout: bool = True,
        dropout_rate: float = 0.0,
        deterministic: bool | None = None,
        precision: PrecisionLike = None,
        kernel_init: Initializer = default_kernel_init,
        out_kernel_init: Initializer | None = None,
        bias_init: Initializer = initializers.zeros_init(),
        out_bias_init: Initializer | None = None,
        use_bias: bool = True,
        attention_fn: Callable[..., Array] = dot_product_attention,
        max_wavelength: float = 10_000.0,
        scale_factor: float = 1.0,
        rngs: rnglib.Rngs,
        keep_rngs: bool = True,
    ):
        self.num_heads = num_heads
        self.num_features = num_features
        self.dtype = dtype
        self.param_dtype = param_dtype
        self.broadcast_dropout = broadcast_dropout
        self.dropout_rate = dropout_rate
        self.deterministic = deterministic
        self.precision = precision
        self.kernel_init = kernel_init
        self.out_kernel_init = out_kernel_init
        self.bias_init = bias_init
        self.out_bias_init = out_bias_init
        self.use_bias = use_bias
        self.attention_fn = attention_fn
        self.max_wavelength = max_wavelength
        self.scale_factor = scale_factor

        if self.num_features % self.num_heads != 0:
            raise ValueError(
                f"Memory dimension ({self.num_features}) must be divisible by "
                f"'num_heads' heads ({self.num_heads})."
            )

        self.head_dim = self.num_features // self.num_heads

        linear_general = functools.partial(
            LinearGeneral,
            out_features=(self.num_heads, self.head_dim),
            dtype=self.dtype,
            param_dtype=self.param_dtype,
            kernel_init=self.kernel_init,
            bias_init=self.bias_init,
            use_bias=self.use_bias,
            precision=self.precision,
        )
        # project inputs_q to multi-headed q/k/v
        # dimensions are then [batch..., length, n_heads, n_features_per_head]
        self.query = linear_general(self.num_features, rngs=rngs)
        self.key = linear_general(self.num_features, rngs=rngs)
        self.value = linear_general(self.num_features, rngs=rngs)

        self.out = LinearGeneral(
            in_features=(self.num_heads, self.head_dim),
            out_features=self.num_features,
            axis=(-2, -1),
            kernel_init=self.out_kernel_init or self.kernel_init,
            bias_init=self.out_bias_init or self.bias_init,
            use_bias=self.use_bias,
            dtype=self.dtype,
            param_dtype=self.param_dtype,
            precision=self.precision,
            rngs=rngs,
        )
        self.rngs = rngs.dropout.fork() if keep_rngs and dropout_rate > 0 else None

    def __call__(
        self,
        inputs: Array,
        *,
        deterministic: bool | None = None,
        rngs: rnglib.Rngs | rnglib.RngStream | None = None,
        positions: Array | None = None,
    ):
        if rngs is None:
            rngs = self.rngs
        elif isinstance(rngs, rnglib.Rngs):
            rngs = rngs.dropout

        if inputs.shape[-1] != self.num_features:
            raise ValueError(
                f"Incompatible input dimension, got {inputs.shape[-1]} "
                f"but module expects {self.num_features}."
            )

        query = self.query(inputs)
        key = self.key(inputs)
        value = self.value(inputs)

        if positions is not None:
            # Apply RoPE to query and key
            query = apply_rope(
                query,
                positions,
                head_dim=self.head_dim,
                max_wavelength=self.max_wavelength,
                scale_factor=self.scale_factor,
            )
            key = apply_rope(
                key,
                positions,
                head_dim=self.head_dim,
                max_wavelength=self.max_wavelength,
                scale_factor=self.scale_factor,
            )

        if self.dropout_rate > 0.0:  # Require `deterministic` only if using dropout.
            deterministic = first_from(
                deterministic,
                self.deterministic,
                error_msg="""No `deterministic` argument was provided to MultiHeadAttention
          as either a __call__ argument, class attribute, or nnx.flag.""",
            )
            if not deterministic:
                if rngs is None:
                    raise ValueError(
                        "'rngs' must be provided to __call__ method if "
                        "MultiHeadAttention instance is defined with keep_rngs=False."
                    )
                dropout_rng = rngs()
            else:
                dropout_rng = None
        else:
            deterministic = True
            dropout_rng = None

        # generate causal mask [1, 1, seq_len, seq_len]
        seq_len = inputs.shape[-2]
        mask = jnp.tril(jnp.ones((seq_len, seq_len), dtype=bool))
        mask = mask[jnp.newaxis, jnp.newaxis, :, :]

        # apply attention
        x = self.attention_fn(
            query,
            key,
            value,
            mask=mask,
            dropout_rng=dropout_rng,
            dropout_rate=self.dropout_rate,
            broadcast_dropout=self.broadcast_dropout,
            deterministic=deterministic,
            dtype=self.dtype,
            precision=self.precision,
            module=None,
        )
        # back to the original inputs dimensions
        out = self.out(x)
        return out



import os
from typing import Dict, Union, List, TypeVar, Generic, Tuple
from safetensors.torch import save_file, load_file, save, load
from soma_models.utils import remap

import torch
import torch.nn as nn

M = TypeVar("M", bound=nn.Module)

# Canonical safetensor key ↔ torch state_dict key
_LAYER_NORM_TO_TORCH: Dict[str, str] = {"gamma": "weight", "beta": "bias"}
_LAYER_NORM_TO_SAFETENSOR: Dict[str, str] = {v: k for k, v in _LAYER_NORM_TO_TORCH.items()}


class Serde(Generic[M]):
    """Symmetric serialize / deserialize between torch modules and safetensors.

    The canonical safetensor format uses:
      - ``gamma`` / ``beta`` for LayerNorm parameters
      - ``(in_features, out_features)`` layout for linear weights  (column-major)
      - flat ``(num_features, num_features)`` layout for attention projection weights

    Torch uses:
      - ``weight`` / ``bias`` for LayerNorm
      - ``(out_features, in_features)`` layout for ``nn.Linear``  (row-major)
    """

    def __init__(self, module: M):
        self.module = module
        self.map_to_torch: Dict[str, str] = {}
        self.map_to_safetensor: Dict[str, str] = {}
        self.linear_weight_keys: set[str] = set()

        for name, child in module.named_modules():
            prefix = name
            if isinstance(child, nn.LayerNorm):
                for st_suffix, torch_suffix in _LAYER_NORM_TO_TORCH.items():
                    st_key = f"{prefix}.{st_suffix}" if prefix else st_suffix
                    torch_key = f"{prefix}.{torch_suffix}" if prefix else torch_suffix
                    self._add(st_key, torch_key)
            elif isinstance(child, nn.Linear):
                key = f"{prefix}.weight" if prefix else "weight"
                self.linear_weight_keys.add(key)

    def _add(self, safetensor_key: str, torch_key: str):
        self.map_to_torch[safetensor_key] = torch_key
        self.map_to_safetensor[torch_key] = safetensor_key

    # -- serialize (torch → safetensor) --

    def _serialize_common(self) -> Dict[str, torch.Tensor]:
        state = self.module.state_dict()
        # Transpose linear weights: torch (out, in) → canonical (in, out)
        for key in self.linear_weight_keys:
            if key in state and state[key].ndim == 2:
                state[key] = state[key].t().contiguous()
        # Rename torch keys → canonical safetensor keys
        remap(state, self.map_to_safetensor, [])
        return dict(sorted(state.items()))

    def serialize(self) -> bytes:
        return save(self._serialize_common())

    def serialize_to_file(self, filename: Union[str, os.PathLike]) -> None:
        return save_file(self._serialize_common(), filename)

    # -- deserialize (safetensor → torch) --

    def _deserialize_common(self, tensors: Dict[str, torch.Tensor]) -> M:
        # Rename canonical keys → torch keys
        remap(tensors, self.map_to_torch, [])
        # Transpose linear weights: canonical (in, out) → torch (out, in)
        for key in self.linear_weight_keys:
            if key in tensors and tensors[key].ndim == 2:
                tensors[key] = tensors[key].t().contiguous()
        self.module.load_state_dict(tensors)
        return self.module

    def deserialize(self, data: bytes) -> M:
        return self._deserialize_common(load(data))

    def deserialize_from_file(self, filename: Union[str, os.PathLike]) -> M:
        return self._deserialize_common(load_file(filename))


def load_safetensor_into(module: nn.Module, data: bytes) -> None:
    """Load safetensors bytes into a torch module (convenience wrapper)."""
    Serde(module).deserialize(data)


class Serializable:
    """Mixin for nn.Module subclasses that adds safetensors serialization.

    Usage::

        class MyModel(Serializable, nn.Module):
            ...

        model.save("weights.safetensors")
        model = MyModel.load("weights.safetensors", ...)
    """

    def save(self, filename: Union[str, os.PathLike]) -> None:
        Serde(self).serialize_to_file(filename)

    def save_bytes(self) -> bytes:
        return Serde(self).serialize()

    @classmethod
    def load(cls, filename: Union[str, os.PathLike], *args, **kwargs):
        module = cls(*args, **kwargs)
        Serde(module).deserialize_from_file(filename)
        return module

    @classmethod
    def load_bytes(cls, data: bytes, *args, **kwargs):
        module = cls(*args, **kwargs)
        Serde(module).deserialize(data)
        return module
